import numpy as np
import time
from scipy.spatial.transform import Rotation
from multiprocessing import Pool
import json

p0=np.array([[1.+0.j,0.],
               [0.,1.]])
p1=np.array([[0.+0.j,1.],
               [1.,0.]])
p2=np.array([[0.+0.j,-1.j],
               [1.j,0.]])
p3=np.array([[1.+0.j,0.],
               [0.,-1.]])

CNOT=np.identity(4)[[0,1,3,2]]
SWAP=np.identity(4)[[0,2,1,3]]

"""every energy is in GHz and time in ns"""

dt=1e-7 #time step in nanoseconds
EZL=4.65 #left Zeeman splitting in GHz
EZR=4.25 #right Zeeman splitting in GHz
aniso = False #If True, both SOI and g tensors are from the experimental fit. Everything is isotropic otherwise
method = 'commensurate' #'original' or 'commensurate', the latter method is faster because it only calculates one cycle of the periodic Hamiltonian

if aniso:
    #experimental B direction as of 06.09.22
    aB,bB=25*(2*np.pi/360),0
    Bext=np.array([np.cos(aB),np.sin(aB)*np.sin(bB),np.sin(aB)*np.cos(bB)])

    #experimental g-tensors; needed only to detemine the rotation due to g-tensor anisotropy, (Zeeman splittings are set later)
    g0L=np.array([[2.3097413, 0.50094669,-0.06031279],
                  [0.50094669, 2.00140483, 0.06610153],
                  [-0.06031279, 0.06610153, 1.5033222]]) 
    g0R=np.array([[1.85601717,-0.57095177, 0.08634011],
                  [-0.57095177, 2.76073622,-0.00585253],
                  [0.08634011,-0.00585253, 1.45916624]])
    evecz=np.array([0,0,1.])

    #calculating left rotation
    BvecL=g0L@Bext/np.linalg.norm(g0L@Bext)
    rotvecL=np.cross(evecz,BvecL)/np.linalg.norm(np.cross(evecz,BvecL))
    rotangL=np.arccos(evecz@BvecL)
    RLt=Rotation.from_rotvec(-rotangL*rotvecL).as_matrix()

    #calculating right rotation
    BvecR=g0R@Bext/np.linalg.norm(g0R@Bext)
    rotvecR=np.cross(evecz,BvecR)/np.linalg.norm(np.cross(evecz,BvecR))
    rotangR=np.arccos(evecz@BvecR)
    RRt=Rotation.from_rotvec(-rotangR*rotvecR).as_matrix()

    #define SO vectors according to the experimental fit
    theta0nlmfit,theta1nlmfit,theta2nlmfit=1.29,1.6297,0.4262
    nso_vec = np.array([np.cos(theta1nlmfit),
                        np.sin(theta1nlmfit)*np.sin(theta2nlmfit),
                        np.sin(theta1nlmfit)*np.cos(theta2nlmfit)])

    #calculating SOI rotation
    RSO=Rotation.from_rotvec(-2*theta0nlmfit*nso_vec).as_matrix()

    #calculating the overall rotation
    Rtilde=RLt@RSO@np.transpose(RRt)
else:
    #everything is isotropic, no net rotation
    Rtilde=np.identity(3)
    
def Fidelities_of_J(J):
    Jzz=J*Rtilde[2,2]
    Jperp=J*(Rtilde[0,0]+Rtilde[1,1]+1.j*(Rtilde[0,1]-Rtilde[1,0]))/2
    EZ=(EZL+EZR)/2
    dEZ=np.sqrt((EZL-EZR)**2+abs(Jperp)**2) #deltaEZ tilde
    nuRabi=abs(Jzz/np.sqrt(15))

    H0 =0.5*EZL*np.kron(p3,p0)+0.5*EZR*np.kron(p0,p3)+sum(
            [J/4*Rtilde[i,j]*np.kron([p1,p2,p3][i],[p1,p2,p3][j]) 
             for i in range(3) 
             for j in range(3)]) #Hamiltonian with exchange in the qubit frame
    vals,_=np.linalg.eig(H0)
    vals.sort()
    omegaac=abs(vals[1]-vals[0]) #resonant drive
    dH=nuRabi*np.kron(p0,p1) #driving Hamiltonian

    if (H0[1,1] - H0[2,2])!=0:
        phi=np.arctan(abs(2*H0[1,2])/(H0[1,1] - H0[2,2]))
    else:
        phi=0
    gamma=np.angle(H0[1,2])
    
    tmax=abs(np.pi/nuRabi/np.cos(phi/2)) #time to complete a pi rotation (in the Bloch sphere sense)

    #calculating the time evolution with the resonant drive
    if method == 'original':
        UCROT=np.identity(4)
        start_time=time.time()
        for ti in range(int(tmax/dt)):
            UCROT = (np.identity(4)-1.j*dt*(H0+dH*np.sin(omegaac*ti*dt)))@UCROT
        print(time.time()-start_time)
    elif method == 'commensurate':
        start_time=time.time()
        Nt_per_cycle = int(2*np.pi/(omegaac*dt)) 
        dt_comm = 2*np.pi/(omegaac*Nt_per_cycle) #time step commensurate with the driving frequency        
        num_cycles = int(tmax/dt_comm)//Nt_per_cycle
        Nt_incomp_cycle = int(tmax/dt_comm)%Nt_per_cycle
        
        UCROT_per_cycle=np.identity(4)    
        for ti in range(Nt_per_cycle):
            UCROT_per_cycle = (np.identity(4)-1.j*dt_comm*(H0+dH*np.sin(omegaac*ti*dt_comm)))@UCROT_per_cycle
        
        UCROT_incomp_cycle=np.identity(4)    
        for ti in range(Nt_incomp_cycle):
            UCROT_incomp_cycle = (np.identity(4)-1.j*dt_comm*(H0+dH*np.sin(omegaac*ti*dt_comm)))@UCROT_incomp_cycle
        
        UCROT=UCROT_incomp_cycle@(np.linalg.matrix_power(UCROT_per_cycle,num_cycles))
        print(time.time()-start_time)
    else:
        print("Method must be either 'original' of 'commensurate'")

    #basis rotation resulting from Jperp
    Urot=np.array([[1,0,0,0],
                   [0,np.cos(phi/2), -np.exp(1.j*gamma)*np.sin(phi/2),0],
                   [0,np.exp(-1.j*gamma)*np.sin(phi/2),np.cos(phi/2),0],
                   [0,0,0,1]]) 

    #assuming that the mixing operation is not possible
    UNCNOT=np.exp(1.j*(EZ-Jzz/4)*tmax)*np.kron(
                p3**(2*np.ceil(abs(-(EZ+0.5*dEZ)*tmax/np.pi-1/2))-(EZ+0.5*dEZ)*tmax/np.pi-1/2),
                p3**(2*np.ceil(abs(-(EZ-0.5*dEZ-Jzz/2)*tmax/np.pi-1/2))-(EZ-0.5*dEZ-Jzz/2)*tmax/np.pi-1/2)
                                                )@UCROT@np.kron(p0,p3**0.5)
    #assuming that the mixing operation is possible
    UNPhiCNOT = np.exp(1.j*(EZ-Jzz/4)*tmax)*np.kron(
                p3**(2*np.ceil(abs(-(EZ+0.5*dEZ)*tmax/np.pi-1/2))-(EZ+0.5*dEZ)*tmax/np.pi-1/2),
                p3**(2*np.ceil(abs(-(EZ-0.5*dEZ-Jzz/2)*tmax/np.pi-1/2))-(EZ-0.5*dEZ-Jzz/2)*tmax/np.pi-1/2)
                                                )@np.transpose(Urot.conj())@UCROT@Urot@np.kron(p0,p3**0.5)

    print(0.25*abs(np.trace(np.transpose(UNCNOT.conj())@UNCNOT)), 0.25*abs(np.trace(CNOT@UNCNOT)), 0.25*abs(np.trace(CNOT@UNPhiCNOT)))
    return [0.25*abs(np.trace(np.transpose(UNCNOT.conj())@UNCNOT)), 0.25*abs(np.trace(CNOT@UNCNOT)), 0.25*abs(np.trace(CNOT@UNPhiCNOT))]

Jlist=np.arange(0.02,1.02,0.02)*(EZL-EZR)
num_cores = 2
if __name__ == '__main__':
    with Pool(num_cores) as p:
        Fiderrors,Fiddirect,Fidphi=np.transpose(p.map(Fidelities_of_J,Jlist))

    with open('Fidelities_isotropic.json', "w") as f:
        f.write(json.dumps([list(Jlist/(EZL-EZR)),list(Fiderrors),list(Fiddirect),list(Fidphi)]))